import os
import ssl
import string
import uuid

import certifi
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

from diqu.utils.log import logger
from diqu.utils.meta import ResultCode


def alert(data, limit: int = 3) -> ResultCode:
    """Slack Alert

    Args:
        data (_type_): DataFrame of test results
        limit (int, optional): Limit the top issues in details of the alert. Defaults to 3.

    Returns:
        ResultCode: Result code
    """
    template_sum = string.Template(
        "🧵 *Summary on $date:*\n\n"
        "  • 🔴 $fail_count failure(s)\n"
        "  • 🟡 $warn_count warning(s)\n"
        "  • 🟢 $pass_count pass(es)\n"
        "  • ⚫ $deperecated_count deprecation(s)"
    )
    template_issue = string.Template("[$index] $issue\n")
    summary = (
        template_sum.substitute(
            date=data["CHECK_TIMESTAMP"].iloc[0],
            fail_count=data[data["TEST_STATUS"] == "fail"].shape[0],
            warn_count=data[data["TEST_STATUS"] == "warn"].shape[0],
            pass_count=data[data["TEST_STATUS"] == "pass"].shape[0],
            deperecated_count=data[data["TEST_STATUS"] == "deprecate"].shape[0],
        )
        if not data.empty
        else "(No Data)"
    )
    issue_data = (
        data[(data["TEST_STATUS"] != "pass") & (data["TEST_STATUS"] != "deprecate")]
        .sort_values("PRIORITY", ascending=True)
        .head(limit)
    )

    if issue_data.empty:
        logger.info("✅ Empty Data | No Alert required > Slack")
        return ResultCode.SUCCEEDED

    issues = ""
    for i in range(len(issue_data)):
        issues += template_issue.substitute(
            index=i + 1,
            issue=issue_data.iloc[i, 0],  # first column: TEST_TITLE
        )

    r = Slack().post_message(
        blocks=[
            {"type": "section", "text": {"type": "mrkdwn", "text": summary}},
            {"type": "divider"},
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"👉 *Top {limit} Issues*:\n\n{issues or '(No Data)'}",
                },
            },
        ]
    )
    if r == ResultCode.SUCCEEDED:
        logger.info("✅ Done > Slack")

    return r


class Slack:
    """Slack Alert class"""

    def __init__(self) -> None:
        """Initialization"""
        self.client = WebClient(
            token=os.environ.get("SLACK_TOKEN"),
            ssl=ssl.create_default_context(cafile=certifi.where()),
        )
        self.channel = os.environ.get("SLACK_CHANNEL")
        self.channel_id = self.find_channel_id(self.channel)

    def post_message(self, text=None, blocks=[]) -> ResultCode:
        """Send Slack message

        Args:
            text (_type_, optional): Simple text. Defaults to None.
            blocks (list, optional): Rich text. Defaults to [].

        Returns:
            ResultCode: Result code
        """
        if not self.channel_id:
            return ResultCode.FAILED

        logger.info(f"Targetted channel: #{self.channel}[{self.channel_id}]")
        sent_blocks = (
            blocks
            if len(blocks) > 0
            else [{"type": "section", "text": {"type": "mrkdwn", "text": text}}]
        )

        message_id = f"mid: #{str(uuid.uuid4()).lower()}"
        sent_blocks.extend(
            [
                {"type": "divider"},
                {
                    "type": "context",
                    "elements": [
                        {"type": "plain_text", "text": message_id, "emoji": True}
                    ],
                },
            ]
        )

        try:
            _ = self.client.chat_postMessage(
                channel=self.channel_id,
                blocks=sent_blocks,
                text=f"{sent_blocks}{message_id}",
                unfurl_links=False,
                unfurl_media=False,
            )
        except SlackApiError as e:
            logger.error(f"Got an error: {e.response['error']}")
            return ResultCode.FAILED

        return ResultCode.SUCCEEDED

    def find_channel_id(self, name: str) -> str:
        """Find Slack channel ID by name

        Args:
            name (str): Channel name

        Returns:
            str: Channel ID
        """
        try:
            for result in self.client.conversations_list():
                for channel in result["channels"]:
                    if channel["name"] == name:
                        return channel["id"]
        except SlackApiError as e:
            logger.error(str(e))
            return None



from pathlib import Path

from pandas import DataFrame

from diqu.sources.base import BaseConnection
from diqu.utils import exception, yml
from diqu.utils.log import logger
from diqu.utils.module import load_module


class SourceFactory:
    """DWH Source Factory class"""

    def __init__(self, **kwargs) -> None:
        """Initilization:
        - Detect the connection configuration in the dbt profile.yml
        - Load DWH module and get the database connection
        """
        self.dbt_project_dir = kwargs.get("project_dir") or Path.cwd()
        self.dbt_profile_dir = kwargs.get("profiles_dir") or (Path.home() / ".dbt")
        self.dbt_profile_name = kwargs.get("profile_name") or yml.load(
            file_path=f"{self.dbt_project_dir}/dbt_project.yml"
        ).get("profile", "diqu")
        self.target = kwargs.get("target")
        self.package = kwargs.get("package")

    def get_connection(self) -> BaseConnection:
        """Get source module's connection

        Returns:
            BaseConnection: Connection object
        """
        dbt_profile_file_path = f"{self.dbt_profile_dir}/profiles.yml"
        with exception.handle_config_errors(
            dbt_profile_file_path,
            f"Profile [{self.dbt_profile_name}] is not found or a fault config",
        ):
            dbt_profile_content = yml.load(file_path=dbt_profile_file_path).get(
                self.dbt_profile_name
            )

            connection_config = dbt_profile_content.get("outputs", {}).get(
                self.target or dbt_profile_content.get("target")
            )

            module_name = connection_config.get("type")
            logger.info(f"Using dbt profile: {self.dbt_profile_name}")
            logger.info(f"Using dbt profiles.yml at: {self.dbt_profile_dir}")
            logger.info(f"Using {module_name} connection")

            if self.package == "csv" and module_name != "csv":
                logger.error(
                    f"csv cannot be using any other source which is: {module_name}"
                )
                return None

            return load_module(name=module_name, package="diqu.sources").get_connection(
                config=connection_config
            )

    def execute(self, query: str) -> DataFrame:
        """Execute SQL query

        Args:
            query (str): SQL query

        Returns:
            DataFrame: Query result
        """
        conn = self.get_connection()
        if conn:
            return conn.execute(query=query)

        return DataFrame()



from diqu.utils.module import load_module


class PackageFactory:
    """Package Factory class"""

    def __init__(self, **kwargs) -> None:
        """Initilization:
        - Load package module
        - Get the configuration
        """
        self.package = load_module(
            name=kwargs.get("package") or "dq_tools", package="diqu.packages"
        )
        self.config = kwargs

    def get_query(self) -> str:
        """Return the SQL query of the configured package

        Returns:
            str: SQL query or csv file name
        """
        return self.package.get_query(config=self.config)



import contextlib

import click


@contextlib.contextmanager
def handle_module_errors(module: str, msg: str = ""):
    try:
        yield
    except Exception as e:
        raise click.UsageError(f"Module {module} could not be found ({msg or str(e)})")


@contextlib.contextmanager
def handle_file_errors(file: str, msg: str = ""):
    try:
        yield
    except Exception as e:
        raise click.FileError(f"File {file} could not be read ({msg or str(e)})")


@contextlib.contextmanager
def handle_config_errors(config: str, msg: str = ""):
    try:
        yield
    except Exception as e:
        raise click.BadParameter(f"{config} is not reachable ({msg or str(e)})")



import os
import abc
import re

from pandas import DataFrame


class SingletonABCMeta(abc.ABCMeta):
    """Singleton Metaclass for ABC"""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    def clear(cls):
        cls._instances = {}


class BaseConnection(metaclass=SingletonABCMeta):
    """Abstract DWH Connection class"""

    def __init__(self, **config) -> None:
        self.config = config

    def get_profile_config(self, attr: str) -> str:
        """Get connection attribute value from profiles.yml file.

        Supporting the detect the ENV_VAR configuration.

        Args:
            attr (str): Attribute name

        Returns:
            str: Attribute value
        """
        attr_value = self.config.get(attr)
        if not attr_value:
            return None

        pattern = r"env_var\('([^']*)'\)"
        match = re.search(pattern, attr_value)
        if match:
            env_var = match.group(1)
            return os.environ.get(str(env_var))

        return attr_value

    @abc.abstractclassmethod
    def execute(self, query: str) -> DataFrame:
        """Execute the SQL query to the DWH and return the DataFrame result

        Args:
            query (str): SQL query
        """



import yaml

from diqu.utils import exception


def load(file_path: str):
    """Load YAML file content

    Args:
        file_path (str): YAML file path

    Raises:
        click.FileError: Read file error

    Returns:
        dict: yaml content
    """
    with exception.handle_file_errors(file_path):
        with open(file_path, "r") as stream:
            return yaml.safe_load(stream)



from importlib import import_module
from types import ModuleType

from diqu.utils import exception


def load_module(name: str, package: str = "diqu") -> ModuleType:
    """Import the module dynamically

    Args:
        name (str): Module name e.g. snowflake, dq_tools, jira

    Raises:
        Exception: Module not found

    Returns:
        ModuleType: Imported module
    """
    with exception.handle_module_errors(name, f"{package}.{name}"):
        return import_module(name=f".{name}", package=package)



from diqu.utils.log import logger
from diqu.utils.module import load_module


class AlertFactory:
    """Alert Factory class"""

    def __init__(self, **kwargs) -> None:
        """Initialization: Load the alert modules"""
        tos = kwargs.get("to") or []
        self.channels = []
        for to in tos:
            try:
                self.channels.append(load_module(name=to, package="diqu.alerts"))
            except Exception as e:
                logger.warning(str(e))

    def is_empty(self) -> bool:
        """Check if any available channels

        Returns:
            bool: True if there is no channels found
        """
        return len(self.channels) == 0

    def run(self, data):
        """Run the alert to the (multiple) channel(s)

        Args:
            data (DataFrame): Issue data collected
        """
        for c in self.channels:
            logger.info(f"Alerting to: {str(c.__name__).split('.')[-1].upper()}")
            c.alert(data=data)



from diqu.alerts.factory import AlertFactory
from diqu.packages.factory import PackageFactory
from diqu.sources.factory import SourceFactory
from diqu.tasks.base import BaseTask
from diqu.utils.meta import ResultCode


class AlertTask(BaseTask):
    """AlertTask class inherited from BaseTask"""

    def __init__(self, **kwargs) -> None:
        """Initilization: Get 3 modules based on the parameters
        1. Package: where to define the Issue query
        2. Source: which DWH connnection to open and to execute the query
        3. Alert: define the alert format, and which channels need to send the alerts
        """
        super().__init__(**kwargs)
        self.alert = AlertFactory(**kwargs)
        if not self.alert.is_empty():
            self.package = PackageFactory(**kwargs)
            self.source = SourceFactory(**kwargs)

    def run(self) -> ResultCode:
        if self.alert.is_empty():
            return ResultCode.SUCCEEDED

        query = self.package.get_query()
        data = self.source.execute(query=query)
        # data.to_csv(".cache/csv__data.csv", index=False)
        return self.alert.run(data=data)



import rich_click as click

from diqu.cli import common
from diqu.tasks.alert import AlertTask


# diqu
@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
    no_args_is_help=True,
    epilog="Specify one of these sub-commands and you can find more help from there.",
)
@click.version_option(common.VERSION)
@click.pass_context
def diqu(ctx, **kwargs):
    """CLI companion tool to support dq-tools package"""


# diqu alert
@diqu.command(name="alert")
@click.pass_context
@common.options
@common.preflight
def alert(ctx, **kwargs):
    """Alert the issues"""
    exit(AlertTask(**kwargs).run())



import os
import functools
import importlib.metadata
from pathlib import Path

import rich_click as click

from diqu.utils.log import logger
from diqu.utils.tracking import track_run

VERSION = importlib.metadata.version("diqu")


def options(func):
    @click.option(
        "--to",
        help="Specify the channel(s) of Alerts. Multiple options.",
        default=["jira"],
        multiple=True,
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--package",
        help="Configure the package which defines the SQL query.",
        default="dq_tools",
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--target",
        "-t",
        help=(
            "Specify dbt target name." "[default: the default target in profiles.yml]"
        ),
        default=None,
        type=click.STRING,
    )
    @click.option(
        "--project-dir",
        help="Specify dbt project dir.",
        default=Path.cwd(),
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--profile-name",
        help=(
            "Specify dbt profile name in dbt_projects.yml file "
            "which will ignore the `--project-dir` value and corresponding parsing stuff"
        ),
        default=None,
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--profiles-dir",
        help="Specify dbt profiles.yml's dir.",
        default=Path.home() / ".dbt",
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--query-dir",
        help="Specify query's directory which is used to scan the query files.",
        default=os.path.abspath(
            f"{os.path.dirname(os.path.realpath(__file__))}/../packages/include"
        ),
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--query-file",
        help=(
            "Specify query file name which contains the query or data. "
            "Default value is vary based on the package module"
        ),
        default=None,
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--query-database",
        help=(
            "Use it when configuring the '$database' in the query. "
            "Assuming the current connection database context if not configured."
        ),
        default=None,
        show_default=True,
        type=click.STRING,
    )
    @click.option(
        "--query-schema",
        help=(
            "Use it when configuring the '$schema' in the query. "
            "Assuming the current connection schema context if not configured."
        ),
        default=None,
        show_default=True,
        type=click.STRING,
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)  # pragma: no cover

    return wrapper


def preflight(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(f"Run with diqu=={VERSION} 🏃")

        # Get Context
        ctx = click.get_current_context()

        # Tracking
        ctx.with_resource(
            track_run(run_command=ctx.command.name, params=ctx.params, version=VERSION)
        )

        return func(*args, **kwargs)  # pragma: no cover

    return wrapper



import os

from diqu.utils import exception
from diqu.utils.log import logger


class Query:
    """Query helper class"""

    def __init__(self, **kwargs) -> None:
        """Initilization:
        - Get query directory
        """
        self.dir = kwargs.get("query_dir")
        self.file = kwargs.get("query_file")
        self.database = kwargs.get("query_database")
        self.schema = kwargs.get("query_schema")
        if not self.dir:
            self.dir = f"{os.path.dirname(os.path.realpath(__file__))}/include"

    def take(self, file: str) -> str:
        """Scan the query directory and get the query string from file

        Args:
            file (str): SQL file name (without the extension `.sql`)

        Raises:
            click.FileError: Read file error

        Returns:
            str: SQL query string
        """
        query_file = f"{self.dir}/{file}"
        logger.info(f"Looking for the query in: {query_file}")
        with exception.handle_file_errors(query_file):
            with open(query_file, "r") as content:
                sql = content.read()
                if self.database:
                    sql = sql.replace("$database", self.database)
                else:
                    sql = sql.replace("$database.", "")  # assume current db

                if self.schema:
                    sql = sql.replace("$schema", self.schema)
                else:
                    sql = sql.replace("$schema.", "")  # assume current schema

                return sql



import os
import re
import string
from datetime import datetime
from typing import Any, List

from jira import JIRA
from pandas import DataFrame

from diqu.utils.log import logger
from diqu.utils.meta import ResultCode


def alert(data) -> ResultCode:
    r = JiraBoard().raise_issues(data=data)
    if r == ResultCode.SUCCEEDED:
        logger.info("✅ Done > JIRA")
    return r


class JiraBoard:
    """JIRA Board management"""

    def __init__(self) -> None:
        """Initialization"""
        self.project_id = os.environ.get("JIRA_PROJECT_ID")
        self.issue_type = os.environ.get("JIRA_ISSUE_TYPE") or "Bug"
        self.conn = self.get_connection()

    def get_connection(self) -> JIRA:
        """Return JIRA Board connection

        Returns:
            JIRA: Connection object
        """
        return JIRA(
            options={"server": os.environ.get("JIRA_SERVER")},
            basic_auth=(
                os.environ.get("JIRA_AUTH_USER"),
                os.environ.get("JIRA_AUTH_PASSWORD"),
            ),
        )

    def raise_issues(self, data: DataFrame) -> ResultCode:
        """Create | Update the Issue issues

        Args:
            data (DataFrame): Input issue data

        Returns:
            ResultCode: Result code
        """
        open_issues_filter = string.Template(
            'project = "$project" '
            'AND statusCategory != "Done" '
            'AND summary ~ "$filter" '
            "ORDER BY created DESC"
        ).substitute(
            project=self.project_id,
            filter=os.environ.get("JIRA_OPEN_ISSUES_FILTER_BY_SUMMARY") or "dq-tools",
        )

        try:
            open_issue_data = self.__get_issues(jql_filter=open_issues_filter)
            joined = open_issue_data.merge(right=data, on="TEST_ID", how="outer")
            self.__create_issues(
                data=joined.loc[
                    (joined["JIRA_ISSUE_KEY"].isnull())
                    & (joined["TEST_STATUS"] != "pass")
                    & (joined["TEST_STATUS"] != "deprecated")
                ]
            )
            self.__update_issues(
                data=joined.loc[
                    joined["JIRA_ISSUE_KEY"].notnull() & joined["TEST_STATUS"].notnull()
                ]
            )
        except Exception as e:
            logger.error(str(e))
            return ResultCode.FAILED

        return ResultCode.SUCCEEDED

    def __build_field_list(self, data: DataFrame) -> List[dict]:
        """Build JIRA issue fields

        Args:
            data (DataFrame): Input as a list of the issues

        Returns:
            List[dict]: List of issue's fields
        """
        data = data.fillna(0)
        description_template = string.Template(
            "h2. Test metadata:\n\n"
            "- *Test ID*: --- $test_id --- \n"
            "- *Latest Status*: $test_status\n\n"
            "- *Latest Run Timestamp*: $check_timestamp (UTC)\n"
            "- *Latest Run Failed Rate*: $failed_rate\n\n"
            "- *Previous statuses*: $prev_statuses\n"
            "- *Previous run timestamps (UTC)*: $prev_check_timestamps\n"
            "- *Previous # of failed records*: $prev_failed_counts\n"
            "- *Previous # of scanned records*: $prev_scanned_counts\n\n"
            "- *tag 1*: $tag_1\n"
            "- *tag 2*: $tag_2\n\n"
            "h2. Managed by diqu | modified at $current_datetime (UTC)"
        )
        return [
            dict(
                description=description_template.substitute(
                    test_id=row["TEST_ID"],
                    test_status=row["TEST_STATUS"],
                    check_timestamp=row["CHECK_TIMESTAMP"],
                    failed_rate=row["FAILED_RATE"],
                    prev_statuses=row["PREV_STATUSES"],
                    prev_check_timestamps=row["PREV_CHECK_TIMESTAMPS"],
                    prev_failed_counts=row["PREV_NO_OF_RECORDS_FAILED"],
                    prev_scanned_counts=row["PREV_NO_OF_RECORDS_SCANNED"],
                    tag_1=row["TAG_1"],
                    tag_2=row["TAG_2"],
                    current_datetime=datetime.utcnow(),
                ),
                summary=row["TEST_TITLE"],
                issuetype=dict(name=self.issue_type),
                project=dict(id=self.project_id),
                labels=[
                    "diqu",
                    f"{str(row['TAG_1']).replace(' ','_').replace('-','_')}",
                    f"{str(row['TAG_2']).replace(' ','_').replace('-','_')}",
                ],
            )
            for _, row in data.iterrows()
        ]

    def __get_issues(self, jql_filter: str = None, limit: int = 100) -> DataFrame:
        """Get open issues in JIRA Board

        Args:
            jql_filter (str, optional): JQL filter on issue summary. Defaults to None.
            limit (int, optional): Specify the number of issues returned. Defaults to 100.

        Returns:
            DataFrame: Frame of Ticket No and Ticket Title
        """
        search_issues = self.conn.search_issues(jql_str=jql_filter, maxResults=limit)
        regexp_str = "--- (.*) ---"  # test_id is wrapped between 2 '---'
        return DataFrame(
            {
                "JIRA_ISSUE_KEY": [str(x.key) for x in search_issues],
                "TEST_ID": [
                    re.search(regexp_str, str(x.fields.description)).group(1)
                    for x in search_issues
                    if re.search(regexp_str, str(x.fields.description)) is not None
                ],
            }
        )

    def __create_issues(self, data: DataFrame) -> Any:
        """Bulk creation of the input issues' data

        Args:
            data (DataFrame): Input issues' data

        Returns:
            Any: None or List of created issues
        """
        if data.empty:
            logger.info("No new issue(s) detected!")
            return None
        logger.info(f"Creating {len(data)} issue(s) ...")
        return self.conn.create_issues(field_list=self.__build_field_list(data=data))

    def __update_issues(self, data: DataFrame) -> Any:
        """Update the existing issue in JIRA Board

        Args:
            data (DataFrame): Input issues' data need updating

        Returns:
            Any:  None or List of updated issues
        """
        if data.empty:
            logger.info("No open issue(s) need updating!")
            return None
        logger.info(f"Updating {len(data)} issue(s) ...")
        results = []
        for _, row in data.iterrows():
            logger.info(f"Updating {row['JIRA_ISSUE_KEY']}...")
            fields_list = self.__build_field_list(
                data=data.loc[data["JIRA_ISSUE_KEY"] == row["JIRA_ISSUE_KEY"]]
            )[0]
            fields_list.pop("issuetype")
            results.append(
                self.conn.issue(row["JIRA_ISSUE_KEY"]).update(fields=fields_list)
            )

        return results



import os
import platform
from contextlib import contextmanager
from datetime import datetime

import pytz
import requests
from snowplow_tracker import Emitter, StructuredEvent, Tracker
from snowplow_tracker import logger as sp_logger

sp_logger.setLevel(60)  # turn off snowplow logging

COLLECTOR_URL = "com-infinitelambda1-saas1.collector.snplow.net"
COLLECTOR_PROTOCOL = "https"


class TimeoutEmitter(Emitter):
    def __init__(self) -> None:
        super().__init__(
            COLLECTOR_URL,
            protocol=COLLECTOR_PROTOCOL,
            buffer_capacity=30,
            on_failure=self.handle_failure,
            method="post",
            byte_limit=None,  # don't set this.
        )

    @staticmethod
    def handle_failure(num_ok, unsent):
        sp_logger.info(f"Failed to track the event: num_ok={num_ok}, unsent={unsent}")

    def _log_request(self, request, payload):
        sp_logger.info(f"Sending {request} request to {self.endpoint}...")
        sp_logger.debug(f"Payload: {payload}")

    def _log_result(self, request, status_code):
        msg = f"{request} request finished with status code: {status_code}"
        if self.is_good_status_code(status_code):
            sp_logger.info(msg)
        else:
            sp_logger.warning(msg)

    def http_post(self, payload):
        self._log_request("POST", payload)
        r = requests.post(
            self.endpoint,
            data=payload,
            headers={"content-type": "application/json; charset=utf-8"},
            timeout=5.0,
        )
        self._log_result("POST", r.status_code)
        return r

    def http_get(self, payload):
        self._log_request("GET", payload)
        r = requests.get(self.endpoint, params=payload, timeout=5.0)
        self._log_result("GET", r.status_code)
        return r


class User:
    def __init__(self) -> None:
        self.do_not_track = True
        self.run_started_at = datetime.now(tz=pytz.utc)

    def state(self):
        return "do not track" if self.do_not_track else "tracking"

    def disable_tracking(self):
        self.do_not_track = True

    def enable_tracking(self):
        self.do_not_track = False


class SpTracker:
    def __init__(self) -> None:
        self.tracker = Tracker(
            namespace="snowplow_tracker",
            app_id="diqu",
            emitters=[TimeoutEmitter()],
        )
        self.user = self.get_user()

    def get_user(self):
        send_anonymous_usage_stats = True
        if os.getenv("DO_NOT_TRACK", "").lower() in ("1", "t", "true", "y", "yes"):
            send_anonymous_usage_stats = False

        if send_anonymous_usage_stats:
            active_user = User()
            try:
                active_user.enable_tracking()
                return active_user
            except Exception:
                return User()

        return User()

    def track(self, **kwargs):
        if self.user.do_not_track:
            return
        else:
            try:
                self.tracker.track(StructuredEvent(**kwargs))
            except Exception:
                return False
        return True

    def get_platform_context(self):
        return {
            "platform": platform.platform(),
            "python": platform.python_implementation(),
            "python_version": platform.python_version(),
        }

    def track_invocation_start(self, invocation_context):
        assert (
            self.user is not None
        ), "Cannot track invocation end when active user is None"
        data = {"invocation_context": invocation_context}
        data.update(self.get_platform_context())

        return self.track(
            category="diqu",
            action="invocation",
            label="start",
            property_=str(data),
        )

    def track_invocation_end(self, invocation_context, result_type=None):
        assert (
            self.user is not None
        ), "Cannot track invocation end when active user is None"
        data = {"invocation_context": invocation_context, "result_type": result_type}
        data.update(self.get_platform_context())

        return self.track(
            category="diqu",
            action="invocation",
            label="end",
            property_=str(data),
        )

    def flush(self):
        try:
            self.tracker.flush()
        except Exception:
            pass


@contextmanager
def track_run(run_command=None, params=None, version=None):
    tracker = SpTracker()
    invocation_context = {
        "command": run_command,
        "params": params,
        "version": version,
    }
    tracker.track_invocation_start(invocation_context)
    try:
        yield
        tracker.track_invocation_end(invocation_context, result_type="ok")
    except Exception:
        tracker.track_invocation_end(invocation_context, result_type="error")
        raise
    finally:
        tracker.flush()



from abc import ABC, abstractclassmethod

from diqu.utils.meta import ResultCode


class BaseTask(ABC):
    """Abstract Task class"""

    def __init__(self, **kwargs) -> None:
        super().__init__()

    @abstractclassmethod
    def run(self) -> ResultCode:
        """Run the task

        Returns:
            ResultCode: Result code
        """



def get_query(config: dict) -> str:
    """Return CSV file name

    Args:
        config (dict): Configuration object

    Returns:
        str: CSV file name
    """
    return Csv(**config).get_query()


class Csv:
    """Package CSV class"""

    def __init__(self, **kwargs) -> None:
        """Initilization"""
        self.query_file = kwargs.get("query_file") or "csv__data.csv"

    def get_query(self) -> str:
        """Return the CSV file name

        Returns:
            str: CSV file name
        """
        return self.query_file



import logging
import time


class LogFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    blue = "\x1b[34;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s  %(name)s: %(levelname)s - %(message)s"

    FORMATS = {
        logging.DEBUG: blue + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%H:%M:%S")
        formatter.converter = time.gmtime
        return formatter.format(record)


logger = logging.getLogger("diqu")
logger.setLevel(logging.DEBUG)
logger.propagate = False

if len(logger.handlers) == 0:
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(LogFormatter())
    logger.addHandler(ch)



from pandas import DataFrame, read_csv

from diqu.sources.base import BaseConnection
from diqu.utils import exception


def get_connection(config: dict) -> BaseConnection:
    """Get the CSV connection

    Args:
        config (dict): Connection attributes

    Returns:
        BaseConnection: CsvConnection
    """
    return CsvConnection(**config)


class CsvConnection(BaseConnection):
    """CSV connection class"""

    def __init__(self, **config) -> None:
        super().__init__(**config)
        self.dir = self.get_profile_config("dir")

    def execute(self, query: str) -> DataFrame:
        """Read csv

        Args:
            query (str): Csv file name

        Returns:
            DataFrame: Csv content
        """
        with exception.handle_file_errors(query):
            return read_csv(f"{self.dir}/{query}")



import snowflake.connector
from pandas import DataFrame

from diqu.sources.base import BaseConnection


def get_connection(config: dict) -> BaseConnection:
    """Get the Snowflake connection

    Args:
        config (dict): Connection attributes

    Returns:
        BaseConnection: SnowflakeConnection
    """
    return SnowflakeConnection(**config)


class SnowflakeConnection(BaseConnection):
    """Snowflake connection class"""

    def __init__(self, **config) -> None:
        super().__init__(**config)

        clean_config = dict(
            account=self.get_profile_config("account"),
            user=self.get_profile_config("user"),
            password=self.get_profile_config("password"),
            role=self.get_profile_config("role"),
            warehouse=self.get_profile_config("warehouse"),
            database=self.get_profile_config("database"),
            schema=self.get_profile_config("schema"),
            session_parameters={"QUERY_TAG": "diqu.sources.snowflake"},
        )

        if self.config.get("authenticator"):
            clean_config.pop("password")
            clean_config["authenticator"] = "externalbrowser"
            clean_config["client_request_mfa_token"] = True

        if self.config.get("private_key") or self.config.get("private_key_path"):
            clean_config.pop("password")
            if self.config.get("private_key"):
                clean_config["private_key"] = self.get_profile_config("private_key")
            else:
                clean_config["private_key_path"] = self.get_profile_config(
                    "private_key_path"
                )
            clean_config["private_key_passphrase"] = self.get_profile_config(
                "private_key_passphrase"
            )

        self.conn = snowflake.connector.connect(**clean_config)

    def execute(self, query: str) -> DataFrame:
        cur = self.conn.cursor()
        try:
            cur.execute(query)
            df_result = cur.fetch_pandas_all()
        finally:
            cur.close()

        return df_result



import os
import string

from diqu.packages.query import Query


def get_query(config: dict) -> str:
    """Return SQL query

    Args:
        config (dict): Configuration object

    Returns:
        str: SQL query string
    """
    return DqTools(**config).get_query()


class DqTools:
    """Package dq-tools class"""

    def __init__(self, **kwargs) -> None:
        """Initilization: Get the query engine"""
        self.query = Query(**kwargs)

    def get_query(self) -> str:
        """Return the SQL query using the query engine

        Returns:
            str: SQL query string
        """
        return string.Template(
            self.query.take(self.query.file or "dq_tools__get_test_results.sql")
        ).substitute(
            filter=os.environ.get("JIRA_OPEN_ISSUES_FILTER_BY_SUMMARY") or "dq-tools",
            deprecated_window_in_days=os.environ.get("ISSUE_DEPRECATED_WINDOW_IN_DAYS")
            or "3",
            update_window_in_days=os.environ.get("ISSUE_UPDATE_WINDOW_IN_DAYS") or "14",
        )
